import java.util.ArrayList;
import java.awt.Toolkit;
import java.io.*; 

/******************************************************************************
*  Arena of battling gladiators (fighters).
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.11
******************************************************************************/

public class Arena {

   //--------------------------------------------------------------------------
   //  Constants
   //--------------------------------------------------------------------------

	final int DEFAULT_NUM_YEARS = 50;
	final int DEFAULT_FIGHTS_PER_YEAR = 24;
   final int DEFAULT_NUM_FIGHTERS = 100;
   final int DEFAULT_PCT_MAGIC_PER_LEVEL = 5;
	final Armor.Type DEFAULT_ARMOR = Armor.Type.Plate;

   //--------------------------------------------------------------------------
   //  Fields
   //--------------------------------------------------------------------------

	/** List of fighters for the arena. */
	Party fighterList;

	/** Number of years to run simulation. */
	int numYears;

	/** Fights held per simulated year. */
	int fightsPerYear;

	/** Desired size of the fighter list. */
	int numFighters;

	/** Starting level for new recruits. */
	int startLevel;

	/** Percent chance of magic boost per level. */
	int pctMagicPerLevel;

	/** Type of fights being arranged. */
	boolean fightManVsMonster;

	/** Apply aging effects? */
	boolean applyAgingEffects;

	/** Report summary statistics? */
	boolean reportFighterStats;

	/** Report full details on each fighter? */
	boolean reportFighterData;

	/** Report kills achieved by each monster type? */
	boolean reportMonsterKills;

	/** Report status at each year-end? */
	boolean reportYearEnd;

	/** Base armor type for fighters. */
	Armor.Type baseArmorType;

	/** Fix primary weapon for fighters. */
	Weapon baseWeaponType;

	/** Age of oldest fighter who ever lived. */
	int supMaxAge;

	/** Flag to escape after parsing arguments. */
	boolean exitAfterArgs;

   //--------------------------------------------------------------------------
   //  Constructors
   //--------------------------------------------------------------------------

   /**
   *  Constructor.
   */
   public Arena () {
		Dice.initialize();
		numYears = DEFAULT_NUM_YEARS;
		fightsPerYear = DEFAULT_FIGHTS_PER_YEAR;
		numFighters = DEFAULT_NUM_FIGHTERS;
		pctMagicPerLevel = DEFAULT_PCT_MAGIC_PER_LEVEL;
		baseArmorType = DEFAULT_ARMOR;
		fighterList = new Party(); 
   }

   /**
   *  Constructor (set mode, size).
   */
	public Arena (boolean fightManVsMonster, int numFighters) {
		this();
		this.fightManVsMonster = fightManVsMonster;
		this.numFighters = numFighters;
	}

   //--------------------------------------------------------------------------
   //  Methods
   //--------------------------------------------------------------------------

	/**
	*  Print usage.
	*/
	public void printUsage () {
		System.out.println("Usage: Arena [options]");
		System.out.println("  where options include:");
		System.out.println("\t-a apply aging effects");
		System.out.println("\t-b base type of armor (=0-3, default 3)");
		System.out.println("\t-f fights per year (default =" + DEFAULT_FIGHTS_PER_YEAR + ")");
		System.out.println("\t-m man-vs-monster (default man-vs-man)");
		System.out.println("\t-n number of men fighting (default =" + DEFAULT_NUM_FIGHTERS + ")");
		System.out.println("\t-p percent chance of magic per level (default =" + DEFAULT_PCT_MAGIC_PER_LEVEL + ")");		
		System.out.println("\t-r reporting types");
		System.out.println("\t\ts summary statistics\ty year-end info");
		System.out.println("\t\td detailed data\t\tk monster kills");
		System.out.println("\t-s start level for fighters (default =0)");		
		System.out.println("\t-x use revised XP award table (from Sup-I)");
		System.out.println("\t-y number of years to simulate (default =" + DEFAULT_NUM_YEARS + ")");
		System.out.println();
	}

	/**
	*  Parse arguments.
	*/
	public void parseArgs (String[] args) {
		for (String s: args) {
			if (s.charAt(0) == '-') {
				switch (s.charAt(1)) {
					case 'a': applyAgingEffects = true; break;
					case 'b': setBaseArmorFromInt(getParamInt(s)); break;
					case 'f': fightsPerYear = getParamInt(s); break; 
					case 'm': fightManVsMonster = true; break;
					case 'n': numFighters = getParamInt(s); break;
					case 'p': pctMagicPerLevel = getParamInt(s); break;
					case 'r': setReportingFromParamCode(s); break;
					case 's': startLevel = getParamInt(s); break;
					case 'x': XPTable.setRevisedXpAwards(true); break;
					case 'y': numYears = getParamInt(s); break; 
					default: exitAfterArgs = true; break;
				}
			}
		}
	}

	/**
	*  Should we exit after parsing arguments?
	*/
	public boolean exitAfterArgs () {
		return exitAfterArgs;
	}

	/**
	*  Get integer following equals sign in command parameter.
	*/
	int getParamInt (String s) {
		if (s.charAt(2) == '=') {
			try {
				return Integer.parseInt(s.substring(3));
			}
			catch (NumberFormatException e) {
			}
		}
		exitAfterArgs = true;
		return -1;
	}

	/**
	*  Set base armor type from integer code.
	*/
	void setBaseArmorFromInt (int code) {
		switch (code) {
			case 0: baseArmorType = null; break;
			case 1: baseArmorType = Armor.Type.Leather; break;
			case 2: baseArmorType = Armor.Type.Chain; break;
			case 3: baseArmorType = Armor.Type.Plate; break;
			default: exitAfterArgs = true; break;
		}	
	}

	/**
	*  Set reporting for param char code.
	*/
	void setReportingFromParamCode (String s) {
		for (int i = 2; i < s.length(); i++) {
			switch (s.charAt(i)) {
				case 's': reportFighterStats = true; break;
				case 'd': reportFighterData = true; break;
				case 'k': reportMonsterKills = true; break;			
				case 'y': reportYearEnd = true; break;
				default: exitAfterArgs = true;			
			}
		}	
	}

   /**
   *  Run the arena's top-level algorithm.
   */
   public void runSim () {
		for (int year = 1; year <= numYears; year++) {
			for (int i = 0; i < fightsPerYear; i++) {
				runOneCycle();		
			}
			yearEnd(year);
		}
   }

   /**
   *  Run one cycle of fights for the whole list.
   */
	public void runOneCycle () {
		recruitNewFighters();
		fighterList.shuffleMembers();
		fightDuels();
		fighterList.bringOutYourDead();
		fighterList.clearFallen();
		fighterList.healAll();
	}

   /**
   *  Fill out the fighter list.
   */
   void recruitNewFighters() {
      while (fighterList.size() < numFighters) {
         fighterList.add(newFighter(startLevel));
      }
   }

   /**
   *  Create a new fighter of the indicated level.
   */
   Character newFighter (int level) {
		String name = NameGenerator.getInstance().getRandom();
      Character f = new Character(name, "Human", "Fighter", level, null); 
		if (baseArmorType != null) {
	      f.setArmor(Armor.makeType(baseArmorType));
		}
		Weapon primaryWeapon = (baseWeaponType == null) ?
			Weapon.randomPrimary() : new Weapon(baseWeaponType);
		f.drawWeapon(primaryWeapon);
		f.addEquipment(Weapon.randomSecondary());
		if (primaryWeapon.getHandsUsed() <= 1) {
	      f.setShield(Armor.makeType(Armor.Type.Shield));
		}
		if (fightManVsMonster) {
	      f.addEquipment(Weapon.silverDagger());
		}
      for (int i = 0; i <= level; i++) {
         f.checkMagicArmsBoost(pctMagicPerLevel);
      }
      return f;
   }

   /**
   *  Fight duels for all fighters in list.
   */
   void fightDuels () {
		if (fightManVsMonster)
			fightDuelsManVsMonster();
		else
			fightDuelsManVsMan();
   }

   /**
   *  Duel each pair of fighters.
   */
   void fightDuelsManVsMan () {
      for (int i = 0; i < fighterList.size() - 1; i += 2) {
			Monster m1 = fighterList.get(i);
			Monster m2 = fighterList.get(i+1);
			FightManager.fight(m1, m2);
			awardXP(m1, m2);
      }
   }

   /**
   *  Duel each fighter against a random monster.
   */
   void fightDuelsManVsMonster () {
      MonsterRandomizer ml = MonsterRandomizer.getInstance();
		for (int i = 0; i < fighterList.size(); i++) {
			Monster fighter = fighterList.get(i);
         Monster monster = ml.randomMonsterOnDungeonLevel(
				Math.max(fighter.getLevel(), 1));
			FightManager.fight(fighter, monster);
			awardXP(fighter, monster);
			if (!monster.horsDeCombat())
				addToKillTally(monster);
      }
   }

   /**
   *  Award XP for victory to winning monster.
   */
	void awardXP (Monster m1, Monster m2) {
		Monster winner = m1.horsDeCombat() ? m2 : m1;
		Monster loser = m1.horsDeCombat() ? m1 : m2;
		awardXP(winner, totalXPAward(loser));
	}

   /**
   *  Award XP and magic to one creature/character.
   */
	void awardXP (Monster m, int xp) {

		// Add XP
      int startLevel = m.getLevel();
      m.addXP(xp);
		
		// Check for level-up
		if (m.getLevel() > startLevel) {
         m.checkMagicArmsBoost(pctMagicPerLevel);
      }
	}

   /**
   *  Compute total XP award from monster + treasure.
   *  See DeltasDnD blog 2014-02-25, 2016-01-11.
   */
	int totalXPAward (Monster monster) {
		XPTable xpt = XPTable.getInstance();
		int monsterXP = monster.getXPAward();
   
      // Linear XP from Vol-1; treasure about 75%
      if (!xpt.useRevisedXpAwards()) {
			return monsterXP * 4;
      }
      
      // Table XP from Sup-I; treasure from 95% to 70%.
      else {
         double treasPercent = 0.95 - 0.01 * (monster.getHD() - 1);
         return (int) Math.round(1/(1 - treasPercent) * monsterXP);
      }
   }

   /**
   *  Add to the kill tally for winning monster type.
   */
   void addToKillTally (Monster monster) {
		Monster prototype = MonsterDatabase.getInstance()
			.getByRace(monster.getRace());
      if (prototype != null) {
      	prototype.addToKillTally(1);
      }
   }

   /**
   *  End the year.
   */
	void yearEnd (int year) {
		for (int i = 0; i < fighterList.size(); i++) {
			Character fighter = (Character) fighterList.get(i);
			fighter.incrementAge(applyAgingEffects);	
			fighter.setHitPointsToMax();
		}		
		reportYearEnd(year);
	}

   /**
   *  Print simulation starting info.
   */
	public void reportStart () {
		System.out.println("ARENA"
			+ ": numFighters " + numFighters
			+ ", numYears " + numYears
			+ ", fights/year " + fightsPerYear 
			+ ", " + (fightManVsMonster ? "man-vs-monster" : "man-vs-man")
			+ "\n");
	}

   /**
   *  Print simulation ending info.
   */
	public void reportEnd () {
		reportFighterStatistics();
		reportFighterData();
		reportMonsterKills();
	}

   /**
   *  Print every individual fighter (for testing small groups).
   */
   void reportFighterData () {
      if (reportFighterData) {
			System.out.println();
			fighterList.sortMembers();
			for (int i = 0; i < fighterList.size(); i++) {
				System.out.println(fighterList.get(i));
         }
         System.out.println();
      }
   }

   /**
   *  Generate and print statistics for the fighter list.
   */
   void reportFighterStatistics () {
	 	if (reportFighterStats) {
	
			//	Create the stat bins
			int maxLevel = fighterList.getMaxLevels();
			StatBin bin[] = new StatBin[maxLevel+1];
			for (int	i = 0; i	<= maxLevel; i++)	{
				bin[i] = new StatBin();
			}			
	
			//	Add all fighters data into	stat bins
			for (int	i = 0; i	< fighterList.size(); i++)	{
				Character fighter	= (Character) fighterList.get(i);
				bin[fighter.getLevel()].addCharacter(fighter);
			}
	
			//	Print	header for statistics
			System.out.println("\nLevel\tNum\tHP\t"
				+ "Str\tInt\tWis\tDex\tCon\tCha\tAge");
	
			//	Print	table	of	bin statistics
			for (int	i = 0; i	<= maxLevel; i++)	{
				if	(bin[i].size()	> 0) {
					System.out.print(i +	"\t");
					bin[i].printAllStats();
				}
			}
			System.out.println();
		}
   }

   /**
   *  Prints number killed by each monster type.
   */
   void reportMonsterKills () {
      if (reportMonsterKills) {
			System.out.println();
         MonsterDatabase ml = MonsterDatabase.getInstance();
         for (int level = 1; level <= 6; level++) {
            System.out.println("Level " + level + " Monsters");
            for (int i = 0; i < ml.size(); i++) {
               Monster mon = ml.get(i);
               if (mon.getTableLevel() == level) {
                  System.out.println(mon.getType() 
                     + ": " + mon.getKillTally());
               }
            }
            System.out.println();
         }
         System.out.println();
      }
   }

   /**
   *  Report year-end summary. 
   */
	void reportYearEnd (int year) {
		if (reportYearEnd) {
			Character oldest = getOldestFighter();
			supMaxAge = Math.max(supMaxAge, oldest.getAge());
			System.out.println("Year " + year
				+ ": max level " + fighterList.getMaxLevels()
				+ ", oldest age " + oldest.getAge()
				+ ", oldest level " + oldest.getLevel()
				+ ", supMaxAge " + supMaxAge);
		}	
	}

   /**
   *  Find oldest fighter the list.
   */
	Character getOldestFighter () {
		Character oldest = null;
		for (int i = 0; i < fighterList.size(); i++) {
			Character fighter = (Character) fighterList.get(i);
			if (oldest == null || fighter.getAge() > oldest.getAge())
				oldest = fighter;
		}
		return oldest;
	}

   /**
   *  Counts fighters at a given level or above.
   */
	public int countMenAboveLevel (int level) {
		return fighterList.countAboveLevel(level);
	}

   /**
   *  Prints fighters at a given level or above.
   */
	public void printMenAboveLevel (int level) {
		fighterList.printAboveLevel(level);
	}

   /**
   *  Set the base armor type.
   */
	public void setBaseArmor (Armor.Type type) {
		baseArmorType = type;
	}

   /**
   *  Set the base weapon type.
   */
	public void setBaseWeapon (Weapon weapon) {
		baseWeaponType = weapon;	
	}

   /**
   *  Main application method.
   */
   public static void main (String[] args) {
		Arena arena = new Arena();
		arena.parseArgs(args);
		if (arena.exitAfterArgs()) {
			arena.printUsage();
		}
		else {
			arena.reportStart();
			arena.runSim();
			arena.reportEnd();
		}
		//Toolkit.getDefaultToolkit().beep();
   }
}

