import java.util.Random;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

/******************************************************************************
*  Dice group for random rolls.
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.3
******************************************************************************/

public class Dice {

	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** Number of dice. */
	private int number;
	
	/** Sides on the dice. */
	private int sides;

	/** Multiplier to the dice (negative = divisor). */
	private int multiplier;
	
	/** Addition to the dice (negative = subtraction). */
	private int addition;

	/** 
	*  Random number generator. 
	*  Maintaining our own generator is more efficient than calls to
	*  the Math routine in a dice-heavy program; (1) we avoid checks
	*  for null random; (2) we can call nextInt directly and avoid 
	*  conversion back-and-forth with double.
	*/
	private static Random random = null;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**	
	*  Constructor (one die only).
	*/
	Dice (int sides) {
		this(1, sides, 1, 0);
	}

	/**	
	*  Constructor (number, sides).
	*/
	Dice (int num, int sides) {
		this(num, sides, 1, 0);
	}

	/**	
	*  Constructor (number, sides, addition).
	*/
	Dice (int num, int sides, int add) {
		this(num, sides, 1, add);
	}

	/**	
	*  Constructor (all fields).
	*/
	Dice (int number, int sides, int mul, int add) {
		this.number = number;
		this.sides = sides;
		this.multiplier = mul;
		this.addition = add;	
	}

	/**	
	*  Constructor (read from string descriptor).
	*
	*  RegEx code from @user1803551 on StackExchange:
	*	http://stackoverflow.com/questions/35020687/
	*  how-to-parse-dice-notation-with-a-java-regular-expression
	*/
	Dice (String s) {
		this(0, 0, 1, 0);
		Pattern p = Pattern.compile(
			"([1-9]\\d*)?d([1-9]\\d*)([/x][1-9]\\d*)?([+-]\\d+)?");
		Matcher m = p.matcher(s);
		if(m.matches()) {
			number = (m.group(1) != null ? Integer.parseInt(m.group(1)) : 1);
			sides = Integer.parseInt(m.group(2));
			if	(m.group(3)	!=	null)	{
				 boolean	positive	= m.group(3).startsWith("x");
				 int val	= Integer.parseInt(m.group(3).substring(1));
				 multiplier	= positive ? val : -val;
			}
			if	(m.group(4)	!=	null)	{
				 boolean	positive	= m.group(4).startsWith("+");
				 int val	= Integer.parseInt(m.group(4).substring(1));
				 addition = positive ?	val :	-val;
			}
		}
		else {
			// Accept a constant number (no "d")
			addition = Integer.parseInt(s);
		}
	}

	/**
	*  Constructor (copy).
	*/
	public Dice (Dice d) {
		this.number = d.number;
		this.sides = d.sides;
		this.multiplier = d.multiplier;
		this.addition = d.addition;	
	}

	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------

	// Basic accessors
	public int getNum () { return number; }
	public int getSides () { return sides; }
	public int getMul () { return multiplier; }
	public int getAdd () { return addition; }

	// Basic mutators
	public void setNum (int num) { number = num; }
	public void setSides (int sides) { this.sides = sides; }
	public void setMul (int mul) { multiplier = mul; }
	public void setAdd (int add) { addition = add; }

	/**	
	*  Initialize the dice random generator.
	*  Must call this before any roll() methods.
	*/
	public static void initialize () {
		random = new Random();
	}

	/**	
	*  Roll one die from a static context.
	*  @return The die-roll.
	*/
	public static int roll (int sides) {
		return random.nextInt(sides) + 1;
	}

	/**	
	*  Rolls the dice.
	*  @return The dice-roll, minimum 1.
	*/
	public int roll () {

		// Roll the dice
		int total = 0;
		for (int i = 0; i < number; i++) {
			total += roll(sides);
		}

		// Apply modifiers
		if (multiplier >= 0) {
			total *= multiplier;
		}
		else {
			total = (total - 1) / (-multiplier) + 1;
		}
		total += addition;
		
		// Return at least 1
		return Math.max(total, 1);
	}

	/**	
	*  Rolls the dice with added modifier.
	*  @return The dice-roll, minimum 1.
	*/
	public int rollPlus (int modifier) {
		return Math.max(1, roll() + modifier);
	}

	/**	
	*  Compute the maximum possible roll.
	*  @return Maximum possible roll.
	*/
	public int maxRoll () {
		int total = number * sides;
		if (multiplier >= 0) {
			total *= multiplier;
		}
		else {
			total = (total - 1) / (-multiplier) + 1;
		}
		total += addition;
		return Math.max(total, 1);
	}

	/**
	*  Identify this object as a string.
	*/
	public String toString() {
		if (number > 0) {
			String s = number + "d" + sides;
			if (multiplier != 1) {
				s += (multiplier >= 0) ?
					"x" + multiplier : "/" + (-multiplier);
			}
			if (addition != 0) {
				s += (addition > 0) ?
					"+" + addition : "" + addition;			
			}
			return s;			
		}
		else {
			return "" + addition;		
		}
	}

	/**
	*  Get a string description of a roll.
	*/
	public String reportRoll () {
		return this + " = " + roll();
	}
	
	/**
	*  Main test function.
	*/
	public static void main (String[] args) {
		Dice.initialize();

		// Test static die roll
		System.out.println("Test Static Rolls");
		System.out.println("d6 = " + Dice.roll(6));
		System.out.println("d20 = " + Dice.roll(20));
		System.out.println();

		// Test numerical constructors
		System.out.println("Test Numeric Constructors");
		System.out.println(new Dice(6).reportRoll());
		System.out.println(new Dice(2, 6).reportRoll());
		System.out.println(new Dice(4, 6, 1).reportRoll());
		System.out.println(new Dice(1, 6, -7).reportRoll());
		System.out.println(new Dice(3, 10, 10, 0).reportRoll());
		System.out.println(new Dice(2, 6, 3, -5).reportRoll());
		System.out.println(new Dice(1, 6, -2, 0).reportRoll());
		System.out.println(new Dice(0, 0, 1, 2).reportRoll());
		System.out.println();
		
		// Test string constructor
		System.out.println("Test String Constructors");
		System.out.println(new Dice("d6").reportRoll());
		System.out.println(new Dice("2d6").reportRoll());
		System.out.println(new Dice("4d6+1").reportRoll());
		System.out.println(new Dice("3d10x10").reportRoll());
		System.out.println(new Dice("2d6x3-5").reportRoll());
		System.out.println(new Dice("d6/2").reportRoll());
		System.out.println(new Dice("2").reportRoll());
		System.out.println();
		
		// Test the division operator for bias
		System.out.println("Test d6/2 for Bias");
		Dice d = new Dice("1d6/2");
		int numOptions = 3;
		int[] count = new int[numOptions];
		int numRolls = 10000;
		for (int i = 0; i < numRolls; i++) {
			count[d.roll() - 1]++;
		}
		for (int i = 0; i < numOptions; i++) {
			System.out.println("Ratio of " + (i+1) + "'s: "
				+ (double) count[i]/numRolls);
		}
		System.out.println();
	}
}

